
class Directory:

    def __init__(self, parent=None):
        self.files = {}
        self._parent = parent
        return

    @property
    def parent(self):
        return self._parent

    def get_this_dir_path(self):
        path = []

        p = self

        while p is not None:
            path.insert(0, p)
            p = p.parent

        return path

    def mkdir(self, name):
        self.files[name] = Directory(self)
        return self.files[name]

    def mkfile(self, name, value=None):
        self.files[name] = File(init_value=value, parent=self)
        return self.files[name]

    def delete(self, name):
        if name in self.files:
            del self.files[name]
        return

    def listdir(self):
        return list(self.files.keys())

    def listdir2(self):
        return self.dirnames(), self.filenames()

    def dirnames(self):
        ret = []
        for i in self.listdir():
            if type(self.files[i]) == Directory:
                ret.append(i)
        return ret

    def filenames(self):
        ret = []
        for i in self.listdir():
            if type(self.files[i]) == File:
                ret.append(i)
        return ret

    def dirs(self):
        for i in self.listdir():
            if type(self.files[i]) == Directory:
                ret.append(self.files[i])
        return ret

    def isdir(self):
        return True

    def isfile(self):
        return False

    def getpath(self, path_lst, create_dirs=False):
        """
        if path is not found and create_dir is True, then the path is created
            as dir
        None returned in case of error
        """
        ret = None

        if len(path_lst) != 0:
            name = path_lst[0]
            if name in self:
                ret = self[name]
            else:
                if create_dirs:
                    ret = self.mkdir(name)
                else:
                    ret = None
        else:
            ret = self

        if ret is not None and len(path_lst) != 0:
            rem_path = path_lst[1:]
            if len(rem_path) == 0:
                pass
            else:
                ret = ret.getpath(rem_path, create_dirs=create_dirs)

        return ret

    def walk(self):
        folders = self.dirnames()
        files = self.filenames()

        path = self.get_this_dir_path()

        yield path, folders, files

        for i in folders:
            for j in self[i].walk():
                yield j
        return

    def __contains__(self, name):
        return name in self.files

    def __getitem__(self, name):
        return self.files[name]


class File:

    def __init__(self, init_value=None, parent=None):
        self._v = init_value
        self._parent = parent
        return

    @property
    def parent(self):
        return self._parent

    def isdir(self):
        return False

    def isfile(self):
        return True

    def get_value(self):
        return self._v

    def set_value(self, v):
        self._v = v
        return




import threading
import logging
import os
import io
import sys
import weakref
import select
import time
import atexit


import wayround_org.utils.path
import wayround_org.utils.stream
import wayround_org.utils.time
import wayround_org.utils.error
import wayround_org.utils.osutils
import wayround_org.utils.weakref
import wayround_org.utils.socket


class LoggingFileLikeObject:

    def __init__(self, log_instance, typ='info'):

        if not typ in ['info', 'error']:
            raise ValueError("invalid typ value")

        self._log = weakref.proxy(log_instance, self._on_log_finalize)
        self._typ = typ
        #self._pipe = os.pipe2(os.O_NONBLOCK)
        self._pipe = os.pipe()
        self._pipe_read_file = open(
            self._pipe[0],
            mode='r',
            # buffering=0,
            closefd=False
            )

        self._sync_out_lock = threading.Lock()

        self._stop_flag = threading.Event()

        _t1 = threading.Thread(
            name="_t1 thread of {}".format(self),
            target=self._thread_run
            )
        _t1.start()

        _t2 = threading.Thread(
            name="_t2 thread of {}".format(self),
            target=self._stop_flag_waiter_thread_func
            )
        _t2.start()

        self._thread = weakref.proxy(_t1)
        self._stop_flag_waiter_thread = weakref.proxy(_t2)

        return

    def __del__(self):
        self.stop()
        return

    def _on_log_finalize(self, ref):
        self.stop()
        return

    def _stop_flag_waiter_thread_func(self):
        while True:
            if self._stop_flag.is_set():
                break
            time.sleep(1)
        threading.Thread(target=self.stop).start()
        return

    def fileno(self):
        return self._pipe[1]

    def close(self):
        # NOTE: leaving this empty. this means what this object's thread
        #       should bot be stopped by file-like .close() method, nor by
        #       sending empty line into file descriptor.
        #       the only mean to sto thread here - should be self._stop_flag
        #       event
        return

    def stop(self):
        self._stop_flag.set()
        return

    def _thread_run(self):

        while True:

            if self._stop_flag.is_set():
                break

            sel_res = select.select([self._pipe_read_file], [], [], 0.5)[0]
            if len(sel_res) == 0:
                continue

            try:
                line = self._pipe_read_file.readline()
            except OSError:
                continue

            with self._sync_out_lock:

                if self._stop_flag.is_set():
                    break

                #line = str(line, 'utf-8')
                line = line.rstrip('\n\r\0')

                if self._typ == 'info':
                    self._log.info(line)

                if self._typ == 'error':
                    self._log.error(line)

        threading.Thread(target=self.stop).start()

        self._pipe_read_file.close()

        try:
            os.close(self._pipe[1])
        except OSError:
            pass
        try:
            os.close(self._pipe[0])
        except OSError:
            pass

        return


class Log:
    """
    Note: this class does not uses stop() method to stop logging. Logging now
          stops automatically with object destruction.
          However, this means what .stdout and .stderr (which may be
          used as parameters for Popen) may be closed before Popen-ed
          application writes all it's stuff into them.

    Note: calling gc.collect() may be required at some points (perhaps at
          program end) to successfuly destroy threading objects created by Log
          for exiting Python interpreter normally.
    """

    def __init__(
            self,
            log_dir,
            logname,
            echo=True,
            timestamp=None,
            longest_logname=None,
            group=None,
            user=None
            ):

        ret = 0
        self.code = 0
        self.fileobj = None
        self.logname = logname
        self.log_filename = None
        self.longest_logname = longest_logname
        self._write_lock = threading.Lock()
        self._stop_lock = threading.Lock()

        # TODO: group and user parameter's behavior need to be improved
        self._group, self._user = wayround_org.utils.osutils.convert_gid_uid(
            group,
            user
            )

        self.stdout = LoggingFileLikeObject(self, 'info')
        self.stderr = LoggingFileLikeObject(self, 'error')

        self.stdout = weakref.proxy(self.stdout)
        self.stderr = weakref.proxy(self.stderr)

        os.makedirs(log_dir, exist_ok=True)

        if (not os.path.isdir(log_dir)
                or os.path.islink(log_dir)
                ):
            logging.error(
                "Current file type is not acceptable: {}".format(
                    log_dir
                    )
                )
            ret = 2

        if ret == 0:
            if self._user is not None and self._group is not None:
                os.chown(
                    log_dir,
                    self._user,
                    self._group
                    )

        if ret == 0:

            if timestamp is None:
                timestamp = wayround_org.utils.time.currenttime_stamp_iso8601()

            filename = wayround_org.utils.path.abspath(
                os.path.join(
                    log_dir,
                    "{ts:26} {name}.txt".format(
                        name=logname,
                        ts=timestamp
                        )
                    )
                )

            self.log_filename = filename

            try:
                self.fileobj = open(filename, 'w')
            except:
                logging.exception("Error opening log file")
                ret = 3
            else:
                self.info(
                    "[{}] log started" .format(
                        self.logname
                        ),
                    echo=echo,
                    timestamp=timestamp
                    )

        if ret == 0:
            if self._user is not None and self._group is not None:
                os.chown(
                    self.log_filename,
                    self._user,
                    self._group
                    )

        if ret != 0:
            raise Exception("Exception during Log creation. read above.")

        return

    def __del__(self):
        self._stop()
        return

    def stop(self, echo=True):
        self._stop(echo=echo)
        return

    def close(self, *args, **kwargs):
        return self.stop(*args, **kwargs)

    def _stop(self, echo=True):

        with self._stop_lock:

            if self.fileobj is not None:
                timestamp = wayround_org.utils.time.currenttime_stamp_iso8601()
                self.info(
                    "[{}] log ended" .format(
                        self.logname
                        ),
                    echo=echo,
                    timestamp=timestamp
                    )

                self.stdout.stop()
                self.stderr.stop()

                self.stdout = None
                self.stderr = None

                self.fileobj.flush()
                self.fileobj.close()
                self.fileobj = None

        return

    def write(self, text, echo=False, typ='info', timestamp=None):

        if not typ in ['info', 'error', 'exception', 'warning']:
            raise ValueError("Wrong `typ' parameter")

        if self.fileobj is None:
            raise Exception("Log output file object is None")

        if timestamp is not None:
            pass
        else:
            timestamp = wayround_org.utils.time.currenttime_stamp_iso8601()

        log_name = self.logname
        if self.longest_logname is not None:
            log_name += ' ' * (self.longest_logname - len(self.logname) - 1)

        icon = 'i'
        if typ == 'info':
            icon = 'i'
        elif typ == 'error':
            icon = 'e'
        elif typ == 'exception':
            icon = 'E'
        elif typ == 'warning':
            icon = 'w'
        else:
            icon = '?'

        msg2 = "[{}] [{:26}] [{}] {}".format(
            icon,
            timestamp,
            log_name,
            text
            )

        msg2_scn = "\033[0;1m[{}] [{:26}] [{}]\033[0m {}".format(
            icon,
            timestamp,
            log_name,
            text
            )

        if icon != 'i':
            msg2_scn = "\033[0;1m[\033[0m\033[0;4m\033[0;5m{}\033[0m\033[0;1m] [{:26}] [{}]\033[0m {}".format(
                icon,
                timestamp,
                log_name,
                text
                )

        with self._write_lock:
            self.fileobj.write(msg2 + '\n')
            self.fileobj.flush()
            if echo:
                sys.stderr.write(msg2_scn + '\n')
                sys.stderr.flush()
        return

    def error(self, text, echo=True, timestamp=None):
        self.write(text, echo=echo, typ='error', timestamp=timestamp)
        return

    def exception(self, text, echo=True, timestamp=None, tb=True):
        # EXCEPTION TEXT:
        ei = wayround_org.utils.error.return_instant_exception_info(
            tb=tb
            )
        ttt = """\
{text}
{ei}
""".format(text=text, ei=ei)

        self.write(ttt, echo=echo, typ='exception', timestamp=timestamp)
        return

    def info(self, text, echo=True, timestamp=None):
        self.write(text, echo=echo, typ='info', timestamp=timestamp)
        return

    def warning(self, text, echo=True, timestamp=None):
        self.write(text, echo=echo, typ='warning', timestamp=timestamp)
        return




import os.path
import hashlib

import yaml

import wayround_org.utils.path


class DataCache:

    def __init__(
            self,
            storage_directory,
            cache_name,
            cache_check_refresh_required_callback,
            cache_check_refresh_required_cb_args,
            cache_check_refresh_required_cb_kwargs,
            refresh_callback,
            refresh_cb_args,
            refresh_cb_kwargs,
            checksum_method
            ):
        """
        storage_directory - directory where cache is stored
        cache_name - name to whcih .cache extension added
        cache_check_refresh_required_callback - callable with two parameters,
            by which DataCache object and complete cache file name is passed.
            this callable should return:
                None - on error, True - if refresh is required, False - if
                refresh is not needed
        refresh_callback - callcble to call whan refresh is needed.
            refresh_callback must accept atleast two parameters: DataCach
            object and complete cache filename. also additional parameters may
            be passed passed by refresh_cb_args and refresh_cb_kwargs. this
            callable should write fresh data into file pointed by first
            parameter
        """

        if not isinstance(storage_directory, str):
            raise TypeError("`storage_directory' must be str")

        if not isinstance(cache_name, str):
            raise TypeError("`cache_name' must be str")

        if not callable(cache_check_refresh_required_callback):
            raise TypeError(
                "`cache_check_refresh_required_callback' must be callable"
                )

        if not callable(refresh_callback):
            raise TypeError("`refresh_callback' must be callable")

        if checksum_method not in ['md5', 'sha1', 'sha256', 'sha512']:
            raise ValueError("invalid `checksum_method' value")

        self.checksum_method = checksum_method

        self.storage_directory = storage_directory

        if not os.path.isdir(storage_directory):
            os.makedirs(storage_directory, exist_ok=True)

        self.cache_name = cache_name

        self.cache_check_refresh_required_callback = \
            cache_check_refresh_required_callback

        self.cache_check_refresh_required_cb_args = \
            cache_check_refresh_required_cb_args

        self.cache_check_refresh_required_cb_kwargs = \
            cache_check_refresh_required_cb_kwargs

        self.refresh_callback = refresh_callback

        self.refresh_cb_args = refresh_cb_args

        self.refresh_cb_kwargs = refresh_cb_kwargs
        return

    def get_complete_filename(self):
        ret = wayround_org.utils.path.join(
            self.storage_directory,
            self.cache_name + '.cache'
            )
        return ret

    def get_complete_filename_cs(self):
        ret = '{}.{}'.format(
            self.get_complete_filename(),
            self.checksum_method
            )
        return ret

    def calc_data_checksum(self):

        ret = None

        filename = self.get_complete_filename()

        if os.path.isfile(filename):

            dig = getattr(hashlib, self.checksum_method)()

            with open(filename, 'br') as f:
                while True:
                    b = f.read(2 * 1024**2)
                    if len(b) == 0:
                        break
                    dig.update(b)
            ret = dig.hexdigest().lower()

        return ret

    def rewrite_actual_checksum(self):
        cs = self.calc_data_checksum()
        cs_fn = self.get_complete_filename_cs()
        if cs is None:
            if os.path.isfile(cs_fn):
                os.unlink(cs_fn)
        else:
            with open(cs_fn, 'w') as f:
                f.write(cs)
        return

    def cache_data_integrity_check(self):
        """
        return:
            True - integrity check passed,
            False - integrity check not passed,
            None - some error
        """

        ret = False

        filename_sum_name = self.get_complete_filename_cs()

        d_c_sum = self.calc_data_checksum()

        if d_c_sum is not None:

            if os.path.isfile(filename_sum_name):
                with open(filename_sum_name) as f:
                    f_sum = f.read()

                f_sum = f_sum.strip().lower()

                if d_c_sum == f_sum:
                    ret = True

        return ret

    def open_cache(self):
        """
        return: None - on error, file opened in 'r' mode - on success
        """
        ret = None
        complete_filename = self.get_complete_filename()
        ct_cb_res = self.cache_check_refresh_required_callback(
            self,
            complete_filename,
            *self.cache_check_refresh_required_cb_args,
            **self.cache_check_refresh_required_cb_kwargs
            )
        if ct_cb_res is None:
            ret = None
        else:
            ct_cb_res = bool(ct_cb_res)
            if ct_cb_res is True:
                self.refresh_callback(
                    self,
                    complete_filename,
                    *self.refresh_cb_args,
                    **self.refresh_cb_kwargs
                    )
            ret = open(complete_filename, 'r')
        return ret


class ShortCSTimeoutYamlCacheHandler:

    def __init__(
            self,
            storage_directory,
            cache_name,
            timeout_delta,
            csmethod,
            freshdata_callback,
            freshdata_callback_args=None,
            freshdata_callback_kwargs=None
            ):
        
        if freshdata_callback_args is None:
            freshdata_callback_args = tuple()

        if freshdata_callback_kwargs is None:
            freshdata_callback_kwargs = dict()

        self.storage_directory = storage_directory
        self.cache_name = cache_name
        self.timeout_delta = timeout_delta
        self.csmethod = csmethod
        self.freshdata_callback = freshdata_callback
        self.freshdata_callback_args = freshdata_callback_args
        self.freshdata_callback_kwargs = freshdata_callback_kwargs

        self.dc = DataCache(
            self.storage_directory,
            self.cache_name,
            self._get_data_cache_check_refresh_required_cb,
            tuple(),
            dict(),
            self._get_data_cache_refresh_cb,
            tuple(),
            dict(),
            self.csmethod
            )
        return

    def _get_data_cache_check_refresh_required_cb(self, cache_mgr, path):

        ret = (
            wayround_org.utils.data_cache_miscs.check_data_cache_timeout(
                path,
                self.timeout_delta
                )
            or not cache_mgr.cache_data_integrity_check()
            )

        return ret

    def _get_data_cache_refresh_cb(self, cache_mgr, path):
        # print("args: {}".format(self.freshdata_callback_args))
        # print("kwargs: {}".format(self.freshdata_callback_kwargs))
        data = self.freshdata_callback(
            *self.freshdata_callback_args,
            **self.freshdata_callback_kwargs
            )
        with open(path, 'w') as f:
            f.write(yaml.dump(data))
        cache_mgr.rewrite_actual_checksum()
        return

    def get_data_cache(self):
        ret = None
        cache = self.dc.open_cache()
        if cache is not None:
            ret = yaml.load(cache.read())
            cache.close()
        return ret




"""
Version comparison utilities
"""

import logging
import os.path
import functools
import copy
import re

import wayround_org.utils.tarball
import wayround_org.utils.path
import wayround_org.utils.directory


def source_version_comparator_keyed():
    return functools.cmp_to_key(source_version_comparator)


def source_version_comparator(
        name1, name2,
        acceptable_source_name_extensions
        ):

    ret = 0

    if isinstance(acceptable_source_name_extensions, str):
        acceptable_source_name_extensions = \
            acceptable_source_name_extensions.split(' ')

    name1 = os.path.basename(name1)
    name2 = os.path.basename(name2)

    d1 = wayround_org.utils.tarball.parse_tarball_name(
        name1,
        True,
        acceptable_source_name_extensions
        )

    d2 = wayround_org.utils.tarball.parse_tarball_name(
        name2,
        True,
        acceptable_source_name_extensions
        )

    if d1 is None:
        raise Exception("Can't parse filename: {}".format(name1))

    if d2 is None:
        raise Exception("Can't parse filename: {}".format(name2))

    if d1['groups']['name'] != d2['groups']['name']:
        raise ValueError(
            "Files has different names: `{}' ({}) and `{}' ({})".format(
                d1['groups']['name'], name1,
                d2['groups']['name'], name2
                )
            )

    else:
        com_res = standard_comparison(
            d1['groups']['version_list'], d1['groups']['status_list'],
            d2['groups']['version_list'], d2['groups']['status_list']
            )

        if com_res != 0:
            ret = com_res
        else:
            ret = 0

    if ret == -1:
        logging.debug(name1 + ' < ' + name2)
    elif ret == 1:
        logging.debug(name1 + ' > ' + name2)
    else:
        logging.debug(name1 + ' = ' + name2)

    return ret


def standard_comparator(
        version1,
        version2
        ):

    logging.debug("standard_comparator: `{}', `{}'".format(version1, version2))

    int_v1 = version1
    int_v2 = version2

    i1_error = False
    i2_error = False

    if isinstance(version1, str):
        int_v1 = version1.split('.')

    if isinstance(version2, str):
        int_v2 = version2.split('.')

    if not isinstance(int_v1, list):
        i1_error = True
    else:
        for i in int_v1:
            if not isinstance(i, (int, str)):
                i1_error = True

    if not isinstance(int_v2, list):
        i2_error = True
    else:
        for i in int_v2:
            if not isinstance(i, (int, str)):
                i2_error = True

    if i1_error:
        raise ValueError(
            "standart_comparison parameters must be [lists of [str or int]]"
            " or [strings], not {}".format(int_v1)
            )

    if i2_error:
        raise ValueError(
            "standart_comparison parameters must be [lists of [str or int]]"
            " or [strings], not {}".format(int_v2)
            )

    ret = standard_comparison(int_v1, None, int_v2, None)
    logging.debug("standard_comparator ret: `{}'".format(ret))
    return ret


def standard_comparison(
        version_list1, status_list1,
        version_list2, status_list2
        ):

    vers_comp_res = None
    stat_comp_res = None

    vers1 = version_list1
    vers2 = version_list2

    longer = None

    v1l = len(vers1)
    v2l = len(vers2)

    #  length used in first comparison part
    el_1 = v1l

    if v1l == v2l:
        longer = None
        el_1 = v1l

    elif v1l > v2l:
        longer = 'vers1'
        el_1 = v2l

    else:
        longer = 'vers2'
        el_1 = v1l

    # first comparison part

    for i in range(el_1):

        if int(vers1[i]) > int(vers2[i]):
            logging.debug(vers1[i] + ' > ' + vers2[i])
            vers_comp_res = +1
            break
        elif int(vers1[i]) < int(vers2[i]):
            logging.debug(vers1[i] + ' < ' + vers2[i])
            vers_comp_res = -1
            break
        else:
            continue

    # second comparison part
    if vers_comp_res is None:
        if longer is not None:
            if longer == 'vers1':
                logging.debug(str(vers1) + ' > ' + str(vers2))
                vers_comp_res = +1
            else:
                logging.debug(str(vers1) + ' > ' + str(vers2))
                vers_comp_res = -1

    if vers_comp_res is None:
        vers_comp_res = 0

    if vers_comp_res == 0:
        if status_list1 is not None and status_list2 is not None:
            s1 = '.'.join(status_list1)
            s2 = '.'.join(status_list2)
            if s1 > s2:
                stat_comp_res = +1
            elif s1 < s2:
                stat_comp_res = -1
            else:
                stat_comp_res = 0

            vers_comp_res = stat_comp_res

    ret = vers_comp_res

    return ret


def same_base_structurize_by_version(bases):
    """
    return example:
    {
        0: ['b-0.tar', 'b-0.0.tar', 'b-0.0.0.tar'],
        1: 'b-1.tar',
        2: {
            0: 'b-2.0.tar',
            1: 'b-2.1.tar'
            2: {
                0: ['b-2.2.0.tar', 'b-2.2.tar', 'b-2.2.0.tar.xz'],
                1: 'b-2.2.1.tar',
                2: 'b-2.2.2.tar'
                },
            3: 'b-2.3.tar'
        }
    }
    """

    ver_tree = wayround_org.utils.directory.Directory()

    for i in bases:
        ver_tree_add_base(ver_tree, i)

    return ver_tree


def restore_base(directory, value, as_name=0):
    if not as_name in directory:
        directory.mkfile(as_name, value)
    else:
        if directory[as_name].isfile():
            fv = directory[as_name].get_value()
            if isinstance(fv, str):
                directory[as_name].set_value([fv])
                fv = directory[as_name].get_value()
            if isinstance(value, list):
                for i in value:
                    fv.append(i)
            else:
                fv.append(value)
            directory[as_name].set_value(list(set(fv)))
        elif directory[as_name].isdir():
            directory = directory[as_name]
            restore_base(
                directory, value,
                as_name=0  # new subversions treated as zeros
                )
        else:
            raise Exception("programming error")
    return


def ver_tree_add_base(ver_tree, base):
    parse_result = wayround_org.utils.tarball.parse_tarball_name(base)

    version_list = parse_result['groups']['version_list']
    version_list_len = len(version_list)

    path_part = version_list[:-1]
    for i in range(len(path_part) - 1, -1, -1):
        path_part[i] = int(path_part[i])

    file_part = int(version_list[-1])

    directory = ver_tree

    if len(path_part) != 0:
        bases = []
        for i in range(len(path_part)):
            ii = path_part[i]
            if ii in directory:
                if directory[ii].isfile():
                    bases.append(directory[ii].get_value())
                    directory.delete(ii)
                    directory.mkdir(ii)
                directory = directory[ii]
            else:
                directory = directory.mkdir(ii)

        for i in bases:
            restore_base(directory, i, as_name=0)

    restore_base(directory, base, file_part)

    return


def truncate_ver_tree(directory, length):
    lst = directory.listdir()
    lst.sort(reverse=True)

    to_delete = lst[length:]
    lst = lst[:length]

    if len(directory.get_this_dir_path()) > 2:
        for i in range(len(to_delete) - 1, -1, -1):
            if re.match(r'9\d+', str(i)):
                lst.append(to_delete[i])
                del to_delete[i]

    for i in to_delete:
        directory.delete(i)

    for i in lst:
        if directory[i].isdir():
            truncate_ver_tree(directory[i], length)

    return


def get_bases_from_ver_tree(directory, preferred_tarball_compressors):
    bases = []
    for path, dirs, files in directory.walk():
        for i in files:
            val = path[-1][i].get_value()
            if not isinstance(val, list):
                val = [val]

            res = wayround_org.utils.tarball.\
                tarball_names_list_subdivide_by_status(
                    val
                )

            for key in res.keys():
                res2 = wayround_org.utils.path.select_by_prefered_extension(
                    res[key],
                    preferred_tarball_compressors
                    )
                bases.append(res2)

            # res = wayround_org.utils.path.select_by_prefered_extension(
            #     val,
            #     preferred_tarball_compressors
            #     )
            # bases.append(res)
            # for j in val:
            #    bases.append(j)
    return bases


def test_same_base_structurize_by_version():

    test_bases = [
        'b-0.tar',
        'b-0.0.tar',
        'b-0.0.0.tar',
        'b-1.tar',
        'b-2.0.tar',
        'b-2.1.tar',
        'b-2.2.0.tar',
        'b-2.2.tar',
        'b-2.2.0.tar.xz',
        'b-2.2.1.tar',
        'b-2.2.2.tar',
        'b-2.3.tar'
        ]

    ver_tree = wayround_org.utils.directory.Directory()

    for i in test_bases:
        ver_tree_add_base(ver_tree, i)

    return ver_tree




import urllib.request

import lxml.html

import wayround_org.utils.directory
import wayround_org.utils.path
import wayround_org.utils.uri


class HTMLWalk:

    def __init__(self, domain, scheme='https', port=None):
        self._scheme = scheme
        self._port = port
        self._domain = domain
        self._tree = wayround_org.utils.directory.Directory()
        self._searched_paths = dict()
        return

    def get_tree(self):
        return self._tree

    def _listdir(self, path, variant=0):

        if isinstance(path, str):
            path = wayround_org.utils.path.split(path.strip('/'))

        if variant == 0:
            ret = None
        elif variant == 1:
            ret = None, None
        else:
            raise Exception("programming error")

        ctpp_res = self.check_tree_path_population(path)
        if ctpp_res is True:
            directory = self._tree.getpath(path)
            if directory is not None:
                if variant == 0:
                    ret = directory.listdir()
                elif variant == 1:
                    ret = directory.listdir2()
                else:
                    raise Exception("programming error")

        return ret

    def listdir(self, path):
        return self._listdir(path, variant=0)

    def listdir2(self, path):
        return self._listdir(path, variant=1)

    def search_objects(self, path_lst):

        if not isinstance(path_lst, list):
            raise TypeError("`path_lst' must be str list")

        path_lst_j = wayround_org.utils.path.join(path_lst)

        if path_lst_j in self._searched_paths:
            ret = self._searched_paths[path_lst_j]
        else:

            ret = None

            port_str = ''
            if self._port is not None:
                port_str = ':{}'.format(self._port)

            uri = '{scheme}://{domain}{port}/{path}'.format(
                scheme=self._scheme,
                domain=self._domain,
                port=port_str,
                path=path_lst_j
                )

            try:
                page = urllib.request.urlopen(uri)
                page_text = page.read()
                page.close()
            except urllib.error.HTTPError:
                page_text = None

            if page_text is not None:
                page = lxml.html.document_fromstring(page_text)

                hrefs = set()

                for i in page.findall('.//a'):
                    hrefs.add(i.get('href', ''))

                hrefs -= set([''])

                ret = list(hrefs)
                ret.sort()

            self._searched_paths[path_lst_j] = ret

        return ret

    def populate_tree_from_search_objects_result(self, path_lst):

        search_objects_res = self.search_objects(path_lst)

        if not isinstance(path_lst, list):
            raise TypeError("`path_lst' must be str list")

        if (search_objects_res is not None
                and not isinstance(search_objects_res, list)):
            raise Exception(
                "Programming error."
                " Here `search_objects_res' must always"
                " be None or list of str"
                )

        if len(path_lst) != 0:
            directory = self._tree.getpath(path_lst, create_dirs=True)
        else:
            directory = self._tree.getpath([])

        if search_objects_res is not None:
            for i in search_objects_res:

                i_unquoted = urllib.request.unquote(i)

                if i_unquoted in ['..', '../', '.', './', '/']:
                    continue

                if i_unquoted.startswith('?'):
                    continue

                if not wayround_org.utils.uri.isuri(i):
                    if i_unquoted.endswith('/') and '/' not in i_unquoted[:-1]:
                        directory.mkdir(i_unquoted[:-1])

                    if (not i_unquoted.endswith('/')
                            and '/' not in i_unquoted[:-1]):
                        directory.mkfile(i_unquoted)

        return

    def check_tree_path_population(self, path_lst):
        """
        return: True - ok, False - error
        """
        if not isinstance(path_lst, list):
            raise TypeError("`path_lst' must be str list")

        path_lst_j = wayround_org.utils.path.join(path_lst)

        if path_lst_j not in self._searched_paths:
            self.search_objects(path_lst)
            if (path_lst_j in self._searched_paths):
                self.populate_tree_from_search_objects_result(path_lst)
        ret = (path_lst_j in self._searched_paths)
        return ret




"""
Here are placed some typical functions for use with data_cahce module tools
"""

import os.path
import logging
import datetime


def check_data_cache_timeout(filename, timeout_delta):
    """
    return:
        True - cache timed out,
        False - cache isn't timed out,
        None - some error
    """

    ret = True

    if not isinstance(filename, str):
        raise TypeError("`filename' must be str")

    if not isinstance(timeout_delta, datetime.timedelta):
        raise TypeError("`timeout_delta' must be datetime.timedelta")

    try:
        if os.path.isfile(filename):

            file_ctime = datetime.datetime.fromtimestamp(
                os.stat(filename).st_ctime,
                tz=datetime.timezone.utc
                )
            current_time = datetime.datetime.now(
                tz=datetime.timezone.utc
                )

            if (current_time - file_ctime) <= timeout_delta:
                ret = False
    except:
        logging.exception('error')
        ret = None

    return ret
